from enum import Enum
from typing import Optional

from .base import BaseGql

from ..api import OAuth2APISession
from ..utils import setup_logger
from ..api import get_app_api_session

LOGGER = setup_logger("st_api")


class GqlAPI(BaseGql):
    class Meta(Enum):
        APP = "ST"
        PATH = "/v2/api/sync"
        TOS2_ENV = "TSS_SERVICE"

    """
    Creates an Oauth2APISession connected to SecureTrack's GraphQL API.   

    By default the value of the "GQL_HOST" environment variable will be used as the hostname. 
    Otherwise, you can provide the hostname argument.

    By default the values of thes "ST_API_USERNAME" and "ST_API_PASSWORD" environment variables 
    will be used for initial oauth2 authentication. Otherwise, you can provide the username and password arguments.

    :param Optional[str] hostname
    :param Optional[str] username
    :param Optional[str] password
    """

    def __init__(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        client_id: Optional[str] = None,
    ):
        super().__init__()

        self.hostname, self.username, self.password, self.session = get_app_api_session(
            app=self,
            hostname=hostname,
            username=username,
            password=password,
            client_id=client_id,
            session_cls=OAuth2APISession,
        )



import json
import pytest
import responses
from requests.exceptions import HTTPError


# from pytos2.securechange.entrypoint import Scw
from pytos2.api import ApiError
from pytos2.securechange.ticket import Ticket, TicketStatus, Task, Comment, Attachment

from pytos2.securechange.user import SCWParty, UserXsiType, UserRole, Permission


class TestEntrypoint:
    @responses.activate
    def test_ticket_search(self, scw, ticket_search_mock):
        arr = scw.ticket_search(status=Task.Status.ASSIGNED)
        # first = next(arr)
        first = arr[0]
        assert first is not None

        # assert first.duration == 2180
        assert first.priority == Ticket.Priority.CRITICAL
        # assert first.workflow_name == "Reg Workflow"
        assert first.id == 8
        assert first.status == TicketStatus.INPROGRESS
        assert first.requester_name == "Smith"
        assert first.subject == "new 1401234982542"
        assert first.sla_status == Ticket.SlaStatus.ESCALATION
        assert first.current_step == "verifier"

    @responses.activate
    def test_ticket_search_by_details(self, scw, ticket_search_by_details_mock):
        arr = scw.ticket_search_by_details(task_status=Task.Status.ASSIGNED)
        # first = next(arr)
        first = arr[0]
        assert first is not None

        assert first.priority == Ticket.Priority.CRITICAL
        # assert first.workflow_name == "Reg Workflow"
        assert first.id == 8
        assert first.status == TicketStatus.INPROGRESS
        assert first.requester_name == "Smith"
        assert first.subject == "new 1401234982542"
        assert first.sla_status == Ticket.SlaStatus.ESCALATION
        assert first.current_step == "verifier"

    @responses.activate
    def test_ticket_search_by_query(self, scw, ticket_search_by_query_mock):
        arr = scw.ticket_search_by_saved_search(saved_search_id=14)
        first = arr[0]
        assert first is not None

        assert first.priority == Ticket.Priority.NORMAL
        assert first.id == 1018
        assert first.status == TicketStatus.INPROGRESS
        assert first.requester_name == "Henry Carr"
        assert first.subject == "CSP - CheckPoint Smart-1 - Palo Alto Prisma "
        assert first.sla_status == Ticket.SlaStatus.OK
        assert first.current_step == "Business Approval"

    @responses.activate
    def test_ticket_search_by_query_failed(
        self, scw, ticket_search_by_query_failed_mock
    ):
        with pytest.raises(ApiError) as exc_info:
            scw.ticket_search_by_saved_search(saved_search_id=140)

        response_json = json.load(
            open("tests/securechange/json/ticket/ticket_search_by_query_id_failed.json")
        )

        assert "400" in str(exc_info.value)
        assert response_json["result"]["message"] == "Query ID 140 was not found."

    @responses.activate
    def test_ticket_search_by_group(self, scw, ticket_search_by_group_mock):
        arr = scw.ticket_search_by_group(group_id=9)
        first = arr[0]
        assert first is not None

        assert first.priority == Ticket.Priority.NORMAL
        assert first.id == 1029
        assert first.status == TicketStatus.INPROGRESS
        assert first.requester_name == "Henry Carr"
        assert first.subject == "AR - Palo Alto - Users from Networks"
        assert first.sla_status == Ticket.SlaStatus.OK
        assert first.current_step == "Business Approval"

    @responses.activate
    def test_ticket_search_by_group_failed(
        self, scw, ticket_search_by_group_failed_mock
    ):
        with pytest.raises(ApiError) as exc_info:
            scw.ticket_search_by_group(group_id=40)

        response_json = json.load(
            open("tests/securechange/json/ticket/ticket_search_by_group_failed.json")
        )

        assert "400" in str(exc_info.value)
        assert response_json["result"]["message"] == "Group ID 40 was not found."

    @responses.activate
    def test_ticket_search_by_free_text(self, scw, ticket_search_by_free_text_mock):
        arr = scw.ticket_search_by_free_text(parameter="taskstatus:ASSIGNED")
        first = arr[0]
        assert first is not None

        assert first.priority == Ticket.Priority.NORMAL
        assert first.id == 1046
        assert first.status == TicketStatus.INPROGRESS
        assert first.requester_name == "Henry Carr"
        assert first.subject == "Test_request_trigger10"
        assert first.sla_status == Ticket.SlaStatus.WARNING
        assert first.current_step == "Test -  Server Decommission Request"

    @responses.activate
    def test_get_users(self, scw, users_mock):
        users = scw.get_users(user_name="r", exact_name=True)

        # user = next(users)
        user = users[0]
        assert user is not None

        assert user.display_name == "r"
        assert user.origin_type == SCWParty.OriginType.LOCAL
        assert user.member_of[0].name == "Regi's Grp"
        assert user.roles[0].name == "Role with WorkFlow"
        assert user.id == 289
        assert user.xsi_type == UserXsiType.USER

    @responses.activate
    def test_get_user(self, scw, users_mock, user_mock):
        user = scw.get_user(45)

        assert user is not None

        assert user.display_name == "Johnny_Smith"
        assert user.origin_type == SCWParty.OriginType.LDAP
        assert user.member_of[0].name == "Advertising"
        assert user.roles[0].name == "Role with WorkFlow"
        assert user.id == 45
        assert user.xsi_type == UserXsiType.USER

        user = scw.get_user(45, expand=True)

        assert user is not None

        assert user.display_name == "johnny_smith"
        assert user.origin_type == SCWParty.OriginType.LDAP
        assert user.member_of[0].name == "Advertising"
        assert user.roles[0].name == "Role with WorkFlow"
        assert user.id == 45
        assert user.xsi_type == UserXsiType.USER

    @responses.activate
    def test_get_attachment(self, scw, all_fields_ticket):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
            json={"ticket": {**all_fields_ticket.data["ticket"], "id": 288}},
        )

        comment = all_fields_ticket.comments[0]
        assert isinstance(comment, Comment)
        attachment = comment.attachments[0]
        assert isinstance(attachment, Attachment)
        assert attachment.uid == "b5672678-d9c5-46ee-87cc-d8fa7fce1a43"
        responses.add(
            responses.GET,
            f"https://198.18.0.1/securechangeworkflow/api/securechange/attachments/{attachment.uid}",
        )
        attachment_content = scw.get_attachment(attachment.uid)
        assert isinstance(attachment_content, bytes)

    @responses.activate
    def test_add_attachment(self, scw):
        id = "b5672678-d9c5-46ee-87cc-d8fa7fce1a43"
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/securechange/attachments",
            id,
        )

        uuid = scw.add_attachment("tests/securechange/files/test.pdf")
        assert isinstance(uuid, str)
        assert uuid == id

    @responses.activate
    def test_add_comment(self, scw):
        id = "1"
        ticket_id = 1
        step_id = 1
        task_id = 1
        responses.add(
            responses.POST,
            f"https://198.18.0.1/securechangeworkflow/api/securechange/tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/comments",
            id,
        )

        comment_id = scw.add_comment(
            ticket_id,
            step_id,
            task_id,
            "New Comment",
            ["b5672678-d9c5-46ee-87cc-d8fa7fce1a43"],
        )
        assert isinstance(comment_id, str)
        assert comment_id == id

    @responses.activate
    def test_delete_comment(self, scw):
        """Delete comment by ticket and comment id"""
        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/404/comments/404",
            status=404,
        )

        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1/comments/1",
            status=200,
        )

        test_one = scw.delete_comment(1, 1)
        test_two = scw.delete_comment(1, "1")
        test_three = scw.delete_comment("1", 1)
        assert all([comment is None for comment in [test_one, test_two, test_three]])

        with pytest.raises(ValueError) as exception:
            scw.delete_comment(404, 404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            scw.delete_comment(404, "404")
        assert "Not Found" in str(exception.value)
        with pytest.raises(ValueError) as exception:
            scw.delete_comment("404", 404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            scw.delete_comment("404", "404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_roles(self, roles_mock, scw):
        roles = scw.get_roles()
        assert roles[0].id == 8
        assert roles[0].name == "Application Admin"
        assert roles[0].description == "Can Approve and Edit SA applications"
        assert roles[0].permissions[0].name == "Assign or reassign tickets to all users"
        assert roles[0].permissions[0].value is False
        assert roles[0].permissions[0].key == "REASSIGN_TICKETS_TO_ALL_USERS"

        role = scw.get_role(8)
        assert role.id == 8
        assert role.name == "Application Admin"
        assert role.description == "Can Approve and Edit SA applications"
        assert role.permissions[0].name == "Assign or reassign tickets to all users"
        assert role.permissions[0].value is False
        assert role.permissions[0].key == "REASSIGN_TICKETS_TO_ALL_USERS"



import pytest
import json

from pytos2.securechange.ticket import Ticket

from pytos2.securechange.fields import MultiAccessRequest, AccessRequest
from pytos2.securechange.risk_results import RiskAnalysisResult, RangeNetworkObject

from netaddr import IPRange


class TestRiskResults:
    @pytest.fixture
    def range_network_object(self) -> RangeNetworkObject:
        j = {"name": "TestRange", "min_ip": "192.168.1.120", "max_ip": "192.168.1.123"}
        return RangeNetworkObject.kwargify(j)

    @pytest.fixture
    def ticket(self):
        j = json.load(open("tests/securechange/json/risk_results_ticket.json"))
        return Ticket.kwargify(j)

    def test_attributes(self, ticket):
        multi_ar = ticket.steps[-1].tasks[0].fields[0]
        assert isinstance(multi_ar, MultiAccessRequest)

        ar: AccessRequest = multi_ar.access_requests[0]
        assert isinstance(ar, AccessRequest)
        result = ar.risk_analysis_result

        assert result.status == RiskAnalysisResult.RiskStatus.HAS_RISK

    def test_range_network_object(self, range_network_object):
        assert range_network_object.ip_range == IPRange(
            "192.168.1.120", "192.168.1.123"
        )



from enum import Enum

from typing import List, Optional

from ..utils import propify, prop
from ..models import Jsonable

from pytos2.securechange.service import PredefinedServiceName


@propify
class Layer7ApplicationService(Jsonable):
    name: Optional[PredefinedServiceName] = prop(None)
    protocol: int = prop()
    min: int = prop()
    max: int = prop()


@propify
class ApplicationIdentity(Jsonable):
    class Prop(Enum):
        APPLICATION_IDENTITY_SERVICES = "application_identity_services"

    name: str = prop()
    services: List[Layer7ApplicationService] = prop(
        factory=list,
        key=Prop.APPLICATION_IDENTITY_SERVICES.value,
        flatify="application_identity_service",
    )



from typing import List

from pytos2.utils import propify, prop
from pytos2.models import Jsonable


@propify
class DeviceExclusions(Jsonable):
    device_ids: List[int] = prop(factory=list, flatify="id")



import json
from typing import List

import pytest  # type: ignore
import responses  # type: ignore
from responses import matchers

from pytos2.securetrack import StAPI
from pytos2.securechange import ScwAPI
from pytos2.securechange.entrypoint import Scw
from pytos2.securechange.ticket import Ticket, ApplicationDetails
from tests.securetrack.conftest import (
    network_objects_mock,
    devices_mock,
    device_rules_mock,
    search_rules_on_open_tickets_mock,
)  # noqa
from pytos2.securechange.fields import MultiAccessRequest

from pytos2.securetrack.device import Device
from pytos2.securetrack.rule import SecurityRule
from pytos2.securetrack.network_object import NetworkObject
from pytos2.securetrack import St

from pytos2.utils import safe_iso8601_date


class MockExit(Exception):
    pass


def _api_path(path):
    return f"https://198.18.0.1/securechangeworkflow/api/securechange/{path}"


@pytest.fixture
def st_api():
    return StAPI(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def st():
    return St(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def device_20_network_objects():
    js = json.load(open("tests/securetrack/json/network_objects/20.json"))
    objects = [
        NetworkObject.kwargify(j) for j in js["network_objects"]["network_object"]
    ]
    return objects


@pytest.fixture
def device_20_rules():
    dev = json.load(open("tests/securetrack/json/devices/device-20.json"))
    dev = Device.kwargify(dev)

    js = json.load(open("tests/securetrack/json/rules/device-20.json"))
    rules = [SecurityRule.kwargify(j) for j in js["rules"]["rule"]]

    for rule in rules:
        rule.device = dev

    return rules


@pytest.fixture
def mock_devices() -> List[Device]:
    js = json.load(open("tests/securetrack/json/devices/devices.json"))
    devices = [Device.kwargify(j) for j in js["devices"]["device"]]
    return devices


@pytest.fixture
def st_devices_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices",
        json=json.load(open("tests/securetrack/json/devices/devices.json")),
    )


@pytest.fixture
def device_20_rules_mock(st, st_devices_mock):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/20/rules",
        json=json.load(open("tests/securetrack/json/rules/device-20.json")),
    )


@pytest.fixture
def device_1_rules():
    dev = json.load(open("tests/securetrack/json/devices/device-1.json"))
    dev = Device.kwargify(dev)

    js = json.load(open("tests/securetrack/json/rules/device-1.json"))
    rules = [SecurityRule.kwargify(j) for j in js["rules"]["rule"]]

    for rule in rules:
        rule.device = dev

    return rules


@pytest.fixture
def all_fields_ticket_json():
    return json.load(open("tests/securechange/json/ticket/all_fields-ticket.json"))


@pytest.fixture
def all_fields_workflow_json(scw_api, all_fields_ticket_json):
    return json.load(open("tests/securechange/json/all_fields-workflow.json"))


@pytest.fixture
def all_fields_step_json(all_fields_ticket_json):
    return [step["name"] for step in all_fields_ticket_json["ticket"]["steps"]["step"]]


@pytest.fixture
def all_fields_field_json(all_fields_ticket_json):
    return all_fields_ticket_json["ticket"]["steps"]["step"][0]["tasks"]["task"][
        "fields"
    ]


@pytest.fixture
def application_details():
    return ApplicationDetails.kwargify(
        json.load(open("tests/securechange/json/application_details.json"))
    )


@pytest.fixture
def post_bad_ticket_json():
    json.load(open("tests/securechange/json/ticket/post_bad_ticket.json"))


@pytest.fixture
def closed_group_modify_ticket(scw_api):
    return Ticket.kwargify(
        json.load(
            open("tests/securechange/json/ticket/closed_group_modify-ticket.json")
        )
    )


@pytest.fixture
def closed_ticket(scw_api):
    return Ticket.kwargify(json.load(open("tests/securechange/json/ticket/redo.json")))


@pytest.fixture
def first_step_ticket(scw_api):
    ticket = json.load(open("tests/securechange/json/ticket/redo.json"))
    redo = ticket["ticket"]
    step_one = redo["steps"]["step"][0]
    redo["status"] = "In Progress"
    redo["steps"]["step"] = [step_one]
    redo["current_step"] = {"id": step_one["id"], "name": step_one["name"]}
    return Ticket.kwargify(ticket)


@pytest.fixture
def redo_ticket(scw_api):
    ticket = json.load(open("tests/securechange/json/ticket/redo.json"))
    redo = ticket["ticket"]
    step_two = redo["steps"]["step"][1]
    redo["status"] = "In Progress"
    redo["steps"]["step"] = redo["steps"]["step"][0:2]
    redo["current_step"] = {"id": step_two["id"], "name": step_two["name"]}
    return Ticket.kwargify(ticket)


@pytest.fixture
def open_access_request_ticket(scw_api):
    return Ticket.kwargify(
        json.load(
            open("tests/securechange/json/ticket/open_with_access_request-ticket.json")
        )
    )


@pytest.fixture
def open_group_modify_ticket(scw_api):
    return Ticket.kwargify(
        json.load(
            open("tests/securechange/json/ticket/open_with_group_modify-ticket.json")
        )
    )


@pytest.fixture
def ar_with_designer_results(scw):
    return MultiAccessRequest.kwargify(
        json.load(
            open(
                "tests/securechange/json/field/multi_access_request_with_designer_results-field.json"
            )
        )
    )


@pytest.fixture
def tickets_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets",
        json=json.load(open("tests/securechange/json/ticket/tickets-0-99.json")),
    )

    responses.add(
        responses.GET,
        "https://10.100.0.1/securechangeworkflow/api/securechange/tickets/fake-url-due-to-params/count_100/start_100",
        json=json.load(open("tests/securechange/json/ticket/tickets-100-199.json")),
    )


@pytest.fixture
def ticket_488():
    return json.load(open("tests/securechange/json/ticket/ticket-488.json"))


@pytest.fixture
def scw_api():
    return ScwAPI(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def scw():
    return Scw(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def ticket_search_mock():
    ticket_search_by_status = json.load(
        open("tests/securechange/json/ticket/ticket_search_assigned.json")
    )

    ticket_search_by_step = json.load(
        open("tests/securechange/json/ticket/ticket_search_by_step_name.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search?status=ASSIGNED",
        json=ticket_search_by_status,
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search?current_step=manual%20step",
        json=ticket_search_by_step,
    )


@pytest.fixture
def ticket_search_by_details_mock():
    ticket_search_by_details_by_status = json.load(
        open("tests/securechange/json/ticket/ticket_search_assigned.json")
    )

    ticket_search_by_details_by_step = json.load(
        open("tests/securechange/json/ticket/ticket_search_by_step_name.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search_by/details?task_status=ASSIGNED",
        json=ticket_search_by_details_by_status,
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search_by/details?current_step=manual%20step",
        json=ticket_search_by_details_by_step,
    )


@pytest.fixture
def ticket_search_by_query_mock():
    ticket_search_by_query_id = json.load(
        open("tests/securechange/json/ticket/ticket_search_by_query_id.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search_by/query?query_id=14",
        json=ticket_search_by_query_id,
        match_querystring=True,
    )


@pytest.fixture
def ticket_search_by_query_failed_mock():

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search_by/query?query_id=140",
        status=400,
        json=json.load(
            open("tests/securechange/json/ticket/ticket_search_by_query_id_failed.json")
        ),
    )


@pytest.fixture
def ticket_search_by_group_mock():
    ticket_search_by_group_id = json.load(
        open("tests/securechange/json/ticket/ticket_search_by_group_id.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search_by/group?group_id=9",
        json=ticket_search_by_group_id,
        match_querystring=True,
    )


@pytest.fixture
def ticket_search_by_group_failed_mock():

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search_by/group?group_id=40",
        status=400,
        json=json.load(
            open("tests/securechange/json/ticket/ticket_search_by_group_failed.json")
        ),
    )


@pytest.fixture
def ticket_search_by_free_text_mock():
    ticket_search_by_free_text = json.load(
        open("tests/securechange/json/ticket/ticket_search_by_free_text.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search_by/free_text?parameter=taskstatus:ASSIGNED",
        json=ticket_search_by_free_text,
        match_querystring=True,
    )


@pytest.fixture
@responses.activate
def all_fields_mock(scw_api, all_fields_ticket_json):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/283",
        json=all_fields_ticket_json,
    )
    return Ticket.get(283)


@pytest.fixture
def all_fields_ticket(scw_api, all_fields_ticket_json):
    return Ticket.kwargify(all_fields_ticket_json)


@pytest.fixture
def dynamic_assignment_ticket(scw_api, all_fields_ticket_json):
    return Ticket.kwargify(
        json.load(open("tests/securechange/json/ticket/dynamic_assignment-ticket.json"))
    )


@pytest.fixture
def get_test_field():
    def f(field_cls):
        xsi_type = [
            a.default for a in field_cls.__attrs_attrs__ if a.name == "xsi_type"
        ][0]
        return field_cls.kwargify(
            json.load(
                open(f"tests/securechange/json/field/{xsi_type.value}-field.json")
            )
        )

    return f


@pytest.fixture
def users_mock():
    users_json = json.load(open("tests/securechange/json/users/users_by_username.json"))
    all_users_json = json.load(open("tests/securechange/json/users/all_users.json"))

    responses.add(
        responses.GET,
        _api_path("users?user_name=r&exact_name=True"),
        match_querystring=True,
        json=users_json,
    )

    responses.add(responses.GET, _api_path("users"), json=all_users_json)


@pytest.fixture
def user_mock():
    user_json = json.load(open("tests/securechange/json/users/user_45.json"))
    group_json = json.load(open("tests/securechange/json/users/user_49.json"))

    responses.add(
        responses.GET, _api_path("users/45"), match_querystring=True, json=user_json
    )

    responses.add(
        responses.GET,
        _api_path("users/49"),
        match=[matchers.query_string_matcher("")],
        json=group_json,
    )


@pytest.fixture
def ticket_history_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
        json=json.load(open("tests/securechange/json/ticket/all_fields-ticket.json")),
    )

    responses.add(
        responses.GET,
        _api_path("tickets/288/history"),
        json=json.load(
            open("tests/securechange/json/ticket_history/ticket_history_288.json")
        ),
    )

    responses.add(
        responses.GET,
        _api_path("tickets/828/history"),
        json=json.load(
            open("tests/securechange/json/ticket_history/ticket_history_828.json")
        ),
    )

    responses.add(
        responses.GET,
        _api_path("tickets/2/history"),
        status=400,
    )

    responses.add(
        responses.GET,
        _api_path("tickets/3/history"),
        body="<xml/>",
    )

    responses.add(
        responses.GET,
        _api_path("tickets/4/history"),
        json={"ticket_history_activities": {}},
    )


@pytest.fixture
def workflow_triggers_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/triggers",
        json=json.load(open("tests/securechange/json/workflow_triggers/triggers.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/triggers?id=4",
        json=json.load(
            open("tests/securechange/json/workflow_triggers/trigger-4.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/triggers?name=Contoso",
        json=json.load(
            open("tests/securechange/json/workflow_triggers/trigger-14.json")
        ),
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/triggers",
        status=201,
    )


@pytest.fixture
def ticket_life_cycle_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
        json=json.load(open("tests/securechange/json/ticket/all_fields-ticket.json")),
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/users",
        json=json.load(open("tests/securechange/json/users/all_users.json")),
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/users/45",
        json=json.load(open("tests/securechange/json/users/user_45.json")),
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/change_requester/45",
        status=200,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/steps/1593/tasks/1625/reassign/45",
        status=200,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/reject?handler_id=45",
        status=200,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/reject",
        status=200,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/cancel?requester_id=45",
        status=200,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/cancel",
        status=200,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/confirm?requester_id=45",
        status=200,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/confirm",
        status=200,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1037",
        json=json.load(
            open("tests/securechange/json/ticket/access_request_for_lifecycle.json")
        ),
    )
    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1037/map_rules?handler_id=5",
        status=201,
    )
    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1037/map_rules",
        status=201,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1037/steps/5689/tasks/5705/fields/61671/designer/redesign",
        status=202,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1037/steps/5689/tasks/5705/fields/61671/designer/device/8/commit",
        status=202,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1037/steps/5689/tasks/5705/fields/61671/designer/device/8/update?force=true",
        status=202,
    )
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1037/steps/5689/tasks/5705/fields/61671/designer/device/8/update",
        status=202,
    )


@pytest.fixture
def ticket_life_cycle_ticket_events_mock():
    def filter_ticket_events_callback(request):
        assert request.params["start"] == "0"
        assert request.params["count"] == "2000"
        assert request.params["get_total"] == "true"

        def test_event(event, filters):
            for key, value in filters:
                if key in ["date_from", "date_to"]:
                    test_value = event["timestamp"]
                    if (key == "date_from" and test_value < value) or (
                        key == "date_to" and test_value > value
                    ):
                        return False
                else:
                    if key in [
                        "assignee_id",
                        "assignee_name",
                        "participant_id",
                        "participant_name",
                    ]:
                        if "task_assigned_data" not in event:
                            return False
                        test_value = event["task_assigned_data"].get(key)
                    else:
                        if key not in ["start", "count", "get_total"]:
                            test_value = event.get(key)
                            if str(test_value) != value:
                                return False
            return True

        json_data = json.load(
            open("tests/securechange/json/ticket_event/ticket_events.json")
        )
        parts = request.url.split("?")
        if len(parts) > 1:
            filters = request.params.items()
            events = []
            for event in json_data["ticket_events"]["ticket_event"]:
                if test_event(event, filters):
                    events.append(event)
            json_data["ticket_events"]["ticket_event"] = events
        return (200, {}, json.dumps(json_data))

    responses.add_callback(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/lifecycle_events",
        callback=filter_ticket_events_callback,
    )
    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/lifecycle_events/historical_events",
        status=202,
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/lifecycle_events/historical_events_status",
        json=json.load(
            open("tests/securechange/json/ticket_event/historical_events_status.json")
        ),
        status=200,
    )


@pytest.fixture
def workflows_mock(scw_api):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/workflows/active_workflows?type=ACCESS_REQUEST",
        json=json.load(
            open("tests/securechange/json/workflows/workflows_ACCESS_REQUEST.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/workflows/active_workflows",
        json=json.load(open("tests/securechange/json/workflows/workflows.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/workflows?name=Fully automated Firewall change request",
        json=json.load(open("tests/securechange/json/workflows/workflow_586.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/workflows?id=586",
        json=json.load(open("tests/securechange/json/workflows/workflow_586.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/workflows?id=600",
        json=json.load(open("tests/securechange/json/workflows/workflow_600.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/workflows?name=Generic Workflow",
        json=json.load(open("tests/securechange/json/workflows/workflow_600.json")),
    )


@pytest.fixture
def extensions_mock():
    responses.add(
        responses.GET,
        _api_path("extensions"),
        json=json.load(open("tests/securechange/json/extension/all_extensions.json")),
    )


@pytest.fixture
def related_rules_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/49/steps/177/tasks/177/fields/2929/related_rules",
        json=json.load(
            open("tests/securechange/json/related_rules/related_rules_results.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/49",
        json=json.load(
            open("tests/securechange/json/ticket/related_rules_ticket.json")
        ),
    )


@pytest.fixture
def roles_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/roles",
        json=json.load(open("tests/securechange/json/role/all_roles.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/roles/8",
        json=json.load(open("tests/securechange/json/role/role_8.json")),
    )


@pytest.fixture
def user_or_group_mock():
    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/users/group",
        status=201,
        headers={
            "Location": "https://198.18.0.1/securechangeworkflow/api/securechange/users/49"
        },
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/users/group/54",
        status=200,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securechangeworkflow/api/securechange/users/70",
        status=201,
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/users/import",
        status=201,
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/users",
        status=201,
        headers={
            "Location": "https://198.18.0.1/securechangeworkflow/api/securechange/users/45"
        },
    )


@pytest.fixture
def saved_search_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries",
        json=json.load(open("tests/securechange/json/saved_search/saved_search.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries/14",
        json=json.load(
            open("tests/securechange/json/saved_search/detailed_saved_search.json")
        ),
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries/",
        status=201,
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries/81",
        status=204,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries/81",
        status=204,
    )


@pytest.fixture
def bad_saved_search_mock():
    responses.add(
        responses.POST,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries/",
        status=400,
        json=json.load(
            open("tests/securechange/json/saved_search/create_bad_saved_search.json")
        ),
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries/97",
        status=400,
        json=json.load(
            open("tests/securechange/json/saved_search/update_bad_saved_search.json")
        ),
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries/97",
        status=400,
        json=json.load(
            open("tests/securechange/json/saved_search/delete_bad_saved_search.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/queries",
        status=401,
        body=open(
            "tests/securechange/json/saved_search/unauthorized_error.json"
        ).read(),
        content_type="text/html",
    )


@pytest.fixture
def domains_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/domains",
        json=json.load(open("tests/securechange/json/domain/all_domains.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/domains/2",
        json=json.load(open("tests/securechange/json/domain/domain_2.json")),
    )


@pytest.fixture
def ticket_requests_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=GROUP_REQUESTS&groupId=9",
        json=json.load(
            open("tests/securechange/json/ticket_requests/group_request-groupId.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=MY_REQUESTS&requestId=1059",
        json=json.load(
            open("tests/securechange/json/ticket_requests/my_requests-requestId.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=ALL_REQUESTS&type=TICKET",
        json=json.load(
            open("tests/securechange/json/ticket_requests/all_request-type.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=MY_REQUESTS&ticketStatus=CLOSED",
        json=json.load(
            open(
                "tests/securechange/json/ticket_requests/my_requests-ticket_status.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=ALL_REQUESTS&ticketId=815",
        json=json.load(
            open("tests/securechange/json/ticket_requests/all_requests-ticketId.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=ALL_REQUESTS&ticketRequiresAttention=true",
        json=json.load(
            open(
                "tests/securechange/json/ticket_requests/all_request-requires_attention.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=MY_REQUESTS&ticketId=1046",
        json=json.load(
            open("tests/securechange/json/ticket_requests/my_request-ticketId.json")
        ),
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/1060/cancel",
        status=200,
    )


@pytest.fixture
def bad_ticket_requests_mock():
    responses.add(
        responses.PUT,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/6000/cancel",
        status=404,
        json=json.load(
            open("tests/securechange/json/ticket_requests/invalid_request_id.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/requests/search?view=blahblah",
        status=404,
        json=json.load(
            open("tests/securechange/json/ticket_requests/bad_request.json"),
        ),
    )



from enum import Enum

import pytest
import requests
import responses
import attr
from requests.exceptions import HTTPError
from traversify import Traverser

from pytos2.api import (
    APISession,
    get_app_api_session,
    resultify_response,
    boolify,
    traversify_response,
)


class TestAPISession:
    @responses.activate
    def test_get_json(self):
        client = APISession(
            hostname="test",
            username="user",
            password="password",
            api_path="api",
            verify=False,
        )
        responses.add(
            responses.GET, "https://test/api/thing", json={"k": "v"}, status=200
        )

        r = client.get("thing")
        assert r.json() == {"k": "v"}

    @responses.activate
    def test_post_json(self):
        client = APISession(
            hostname="test",
            username="user",
            password="password",
            api_path="api",
            verify=False,
        )
        responses.add(
            responses.POST, "https://test/api/thing", json={"k": "v"}, status=200
        )

        p = client.post("thing", json={})
        assert p.json()["k"] == "v"


class TestGetAppSession(object):
    @attr.s(auto_attribs=True)
    class App:
        class Meta(Enum):
            APP = "app"
            PATH = "/app"
            TOS2_ENV = "TSS_SERVICE"

    app = App()

    def test_with_args(self):
        hostname, username, password, session = get_app_api_session(
            self.app, "hostname", "username", "password"
        )
        assert hostname == "hostname"
        assert username == "username"
        assert password == "password"
        assert session.base_url.startswith("https:")

    def test_with_env(self, monkeypatch):
        monkeypatch.setenv("APP_API_USERNAME", "env_username")
        monkeypatch.setenv("APP_API_PASSWORD", "env_password")
        monkeypatch.setenv("APP_HOSTNAME", "198.18.0.2")

        hostname, username, password, session = get_app_api_session(self.app)
        assert hostname == "198.18.0.2"
        assert username == "env_username"
        assert password == "env_password"
        assert session.base_url.startswith("https:")
        assert not session.verify

    def test_with_tos2_env(self, monkeypatch):
        monkeypatch.setenv("APP_API_USERNAME", "env_username")
        monkeypatch.setenv("APP_API_PASSWORD", "env_password")
        monkeypatch.setenv("TSS_SERVICE_HOST", "198.18.0.3")
        monkeypatch.setenv("TSS_SERVICE_PORT", "80")

        hostname, username, password, session = get_app_api_session(self.app)
        assert hostname == "198.18.0.3"
        assert username == "env_username"
        assert password == "env_password"
        assert session.base_url.startswith("http:")
        assert not session.verify

    def test_with_missing_args(self):
        with pytest.raises(ValueError) as e:
            get_app_api_session(self.app)
        assert "hostname argument must be provided" in str(e.value)
        with pytest.raises(ValueError) as e:
            get_app_api_session(self.app, "hostname")
        assert "username argument must be provided" in str(e.value)
        with pytest.raises(ValueError) as e:
            get_app_api_session(self.app, "hostname", "password")
        assert "password argument must be provided" in str(e.value)


@responses.activate
def test_resultify_response():
    client = APISession(hostname="test", username="", password="")
    responses.add(responses.GET, "https://test/thing", json={"k": "v"}, status=200)
    responses.add(responses.GET, "https://test/thong", json={"k": "v"}, status=404)

    r = resultify_response(client.get("thing"))
    assert r and isinstance(r, requests.Response)

    with pytest.raises(HTTPError):
        resultify_response(client.get("thong"))
    # assert not r.is_ok() and isinstance(r.value, requests.Response)


@responses.activate
def test_traversify_response():
    client = APISession(hostname="test", username="", password="")
    responses.add(responses.GET, "https://test/thing", json={"k": "v"}, status=200)
    responses.add(responses.GET, "https://test/thong", json={"k": "v"}, status=404)

    r = traversify_response(client.get("thing"))
    assert r and isinstance(r, Traverser)

    with pytest.raises(HTTPError):
        traversify_response(client.get("thong"))


def test_boolify():
    assert boolify(
        {"True": True, "False": False, "None": None, "ignore": "ignore"}
    ) == {"True": "true", "False": "false", "None": None, "ignore": "ignore"}



import pytest
import json
import responses

from pytos2.securetrack.application import (
    ObjectReference,
    Application,
    SingleApplication,
    ApplicationGroup,
)
from pytos2.utils import get_api_node


class TestApplications:
    @responses.activate
    def test_get_applications(self, applications_mock, devices_mock, st):
        applications = st.get_device_applications(174)
        assert len(applications) == 3

        assert isinstance(applications[0], SingleApplication)
        assert applications[0].id == 742402
        assert applications[0].name == "Any"
        assert applications[0].display_name == "Any"
        assert applications[0].class_name == "any_object"
        assert applications[0].type == "any"
        assert applications[0].comment == ""
        assert applications[0].uid == "{bf62fdcf-e2c8-442e-a455-8bbc070e1cc3}"
        assert applications[0].overrides is False
        assert applications[0].management_domain == ""
        assert applications[0].global_device is None
        assert applications[0].device_id is None
        assert applications[0].application_id == ""
        assert applications[0].device_name == ""
        assert len(applications[0].services) == 0

        assert isinstance(applications[1], SingleApplication)
        assert applications[1].id == 742405
        assert applications[1].name == "104apci-unnumbered"
        assert applications[1].display_name == "104apci-unnumbered"
        assert applications[1].class_name == "application_object"
        assert applications[1].type == "single"
        assert applications[1].comment == ""
        assert applications[1].uid == "{286C3A6E-463C-5E3E-3427-4A98110EC5D6}"
        assert applications[1].overrides is False
        assert applications[1].management_domain == "Predefined"
        assert applications[1].global_device is None
        assert applications[1].device_id is None
        assert applications[1].application_id == ""
        assert applications[1].device_name == ""
        assert len(applications[1].services) == 1

        assert isinstance(applications[1].services[0], ObjectReference)
        assert applications[1].services[0].id == 6295334
        assert (
            applications[1].services[0].uid == "{3E95B1FB-7F0F-FE14-4DB7-CD54836E849B}"
        )
        assert applications[1].services[0].name == "st_implicit_app_100bao_2"
        assert applications[1].services[0].display_name == "st_implicit_app_100bao_2"
        assert applications[1].services[0].management_domain == "Predefined"
        assert applications[1].services[0].type == ""
        assert len(applications[1].services[0].ips) == 0
        assert applications[1].services[0].link == ""
        assert len(applications[1].services[0].members) == 0

        assert isinstance(applications[2], ApplicationGroup)
        assert applications[2].id == 744449
        assert applications[2].name == "meebo"
        assert applications[2].display_name == "meebo"
        assert applications[2].class_name == "application_group"
        assert applications[2].type == "group"
        assert applications[2].comment == ""
        assert applications[2].uid == "{69C3839D-B9F8-090D-2FB2-0B0A0672F761}"
        assert applications[2].overrides is False
        assert applications[2].management_domain == "Predefined"
        assert applications[2].global_device is None
        assert applications[2].device_id is None
        assert applications[2].application_id == ""
        assert applications[2].device_name == ""
        assert len(applications[2].applications) == 1

        assert isinstance(applications[2].applications[0], ObjectReference)
        assert applications[2].applications[0].id == 744451
        assert (
            applications[2].applications[0].uid
            == "{5C51B093-3B98-22BE-16B6-3A1F879699D1}"
        )
        assert applications[2].applications[0].name == "meebo-file-transfer"
        assert applications[2].applications[0].display_name == "meebo-file-transfer"
        assert applications[2].applications[0].management_domain == "Predefined"
        assert applications[2].applications[0].type == ""
        assert len(applications[2].applications[0].ips) == 0
        assert applications[2].applications[0].link == ""
        assert len(applications[2].applications[0].members) == 0

        applications = st.get_device_applications("SGW_200.237")
        assert applications == []

    @responses.activate
    def test_get_applications_failed(self, applications_failed_mock, devices_mock, st):
        with pytest.raises(ValueError) as e:
            st.get_device_applications(1337)

        assert "Failed to decode" in str(e.value)

        with pytest.raises(ValueError) as e:
            st.get_device_applications(1338)

        assert "Failed to get resource" in str(e.value)

        with pytest.raises(ValueError) as e:
            st.get_device_applications("NONEXISTENT")

        assert "Cannot find device" in str(e.value)



from dotenv import dotenv_values
from types import List
from pytos2.securechange import Scw as SecureChange
from pytos2.securechange.trigger import on_advance, step_is
from pytos2.securechange.ticket import Ticket, Comment

config = dotenv_values(".env")

HOST = "" or config["AURORA_HOST"]
USER = "" or config["USER"]
PASS = "" or config["PASS"]

# Instantiate Class
secure_change = SecureChange(HOST, USER, PASS)

STEP_NAME = "Trigger Test"


@on_advance(when=step_is(STEP_NAME))
def trigger_test(ticket: Ticket):
    ticket.add_comment(f"Tufin Ticket #{ticket.id}")
    ticket.advance()


def main():
    # Dummy function for console script to use
    pass



from enum import Enum
from typing import Optional, List, Union

from netaddr import IPRange, IPAddress, IPNetwork

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, stringify_optional_obj

from attr.converters import optional


@propify
class NetworkObject(Jsonable):
    class XsiType(Enum):
        HOST_NETWORK_OBJECT = "ns_sc_policy:host_network_object"
        HOST_NETWORK_OBJECT_NO_NAMESPACE = "host_network_object"
        HOST_NETWORK_OBJECT_WITH_INTERFACES = "host_network_object_with_interfaces"
        NETWORK_OBJECT_GROUP = "network_object_group"
        CLOUD_SECURITY_GROUP = "cloud_security_group"
        SUBNET_NETWORK_OBJECT = "subnet_network_object"
        RANGE_NETWORK_OBJECT = "range_network_object"
        ANY_NETWORK_OBJECT = "any_network_object"

    class ClassName(Enum):
        ADDRESS_RANGE = "address_range"
        MULTICAST_ADDRESS_RANGE = "multicast_address_range"
        VPN_ROUTE = "vpn_route"
        INTERFACE = "interface"
        NETWORK_INTERFACE = "network_interface"
        FIREWALL_POLICY = "firewall_policy"
        SUB_POLICY = "sub_policy"
        SECURITY_RULE = "security_rule"
        ROUTING_ENTRY = "routing_entry"
        SECURITY_RULES_GROUP = "security_rules_group"
        RULE_SOURCE = "rule_source"
        RULE_DESTINATION = "rule_destination"
        RULE_INSTALL = "rule_install"
        OTHER_SERVICE = "other_service"
        OTHER_OBJECT = "other_object"
        ACCEPT_ACTION = "accept_action"
        DROP_ACTION = "drop_action"
        DENY_ACTION = "deny_action"
        CONTINUE_ACTION = "continue_action"
        TRUST_ACTION = "trust_action"
        RESET_CLIENT_ACTION = "reset_client_action"
        RESET_SERVER_ACTION = "reset_server_action"
        RESET_BOTH = "reset_both_action"
        NETWORK_OBJECT_GROUP = "network_object_group"
        GROUP_WITH_EXCEPTION = "group_with_exception"
        NETWORK = "network"
        HOST = "host_plain"
        FQDN = "fqdn"
        VIRTUAL_MACHINE_INSTANCE = "vm_instance"
        CONTROLLER = "controllers"
        TCP_SERVICE = "tcp_service"
        UDP_SERVICE = "udp_service"
        SCTP_SERVICE = "sctp_service"
        ICMP_SERVICE = "icmp_service"
        INSTALL_ON = "install_on"
        IP_SERVICE = "ip_service"
        SECURITY_HEADER_RULE = "security_header_rule"
        ZONE = "zone"
        DOMAIN = "domain"
        INTERNET_SERVICE = "internet-service"
        IPV6 = "ipv6_object"
        Range = "address_range"
        SERVICE_GROUP = "service_group"
        POLICIES_COLLECTION = "policies_collection"
        REJECT_ACTION = "reject_action"
        USER_AUTHENTICATE = "user_authenticate"
        CLIENT_AUTHENTICATE = "client_authenticate"
        UNSUPPORTED_NETWORK_OBJECT = "unsupported_object"
        SCHEDULED_EVENT = "scheduled_event"
        TIME = "time"
        TIME_PERIOD = "time_period"
        NAT_POOL = "nat_pool"
        NAT_HEADER = "nat_header_rule"
        ADTR_TRANSLATION_RULE = "address_translation_rule"
        TRANSLATE_STATIC = "translate_static"
        TRANSLATE_HIDDEN = "translate_hidden"
        TRANSLATE_SERVICE = "service_translate"
        BEHIND_INTERFACE_NAT = "behind_interface_nat"
        SECURITY_PROFILE = "security_profile"
        SECURITY_PROFILE_GROUP = "security_profile_group"
        RULE_SERVICES = "rule_services"
        RULE_VPN = "rule_vpn"
        ICMPV6_SERVICE = "icmpv6_service"
        ANY_OBJECT = "any_object"
        ACCESS_ROLE = "access_role"
        USER = "user"
        USER_GROUP = "user_group"
        AD_BRANCH = "ad_branch"
        AD_GROUP = "ad_group"
        AD_USER = "ad_user"
        ALL_IDENTIFIED = "all_identified"
        ENCRYPT = "encrypt"
        CLIENT_ENCRYPT = "client_encrypt"
        RPC_SERVICE = "rpc_service"
        DCE_RPC_SERVICE = "dcerpc_service"
        UNSUPPORTED_SERVICE = "unsupported_service"
        INLINE_LAYER_ACTION = "inline_layer_action"
        SUB_POLICY_COLLECTION = "sub_policy_collection"
        DYNAMIC_OBJECT = "dynamic_object"
        HOST_CKP = "host_ckp"
        GATEWAY_CKP = "gateway_ckp"
        SOFTWARE_GATEWAY = "sofaware_gateway"
        EMBEDDED_DEVICE = "embedded_device"
        GATEWAY_CLUSTER = "gateway_cluster"
        VS_CLUSTER_NETOBJ = "vs_cluster_netobj"
        VSX_CLUSTER_NETOBJ = "vsx_cluster_netobj"
        CLUSTER_MEMBER = "cluster_member"
        GPRS_APN = "gprs_apn"
        UTM_CLUSTER_MEMBER = "utm_cluster_member"
        VS_CLUSTER_MEMBER = "vs_cluster_member"
        VSX_CLUSTER_MEMBER = "vsx_cluster_member"
        VS_NETOBJ = "vs_netobj"
        VSX_NETOBJ = "vsx_netobj"
        SECURITY_ZONE = "security_zone"
        GATEWAY_PLAIN = "gateway_plain"
        GSN_HANDOVER_GROUP = "gsn_handover_group"
        DATA_CENTER = "data_center"
        DATA_CENTER_OBJECT = "data_center_object"
        ENDPOINT_GROUP = "endpoint_group"
        BRIDGE_DOMAIN = "bridge_domain"
        LEARNED_ENDPOINT = "learned_endpoint"
        EXTERNAL_ROUTED_NETWORK = "external_routed_network"
        LOGICAL_NODE_PROFILE = "logical_node_profile"
        LOGICAL_INTERFACE_PROFILE = "logical_interface_profile"
        EXTERNAL_EPG = "external_endpoint_group"
        CONFIGURED_NODE = "configured_node"
        EXTERNAL_INTERFACE = "external_interface"
        EPGS_GROUP = "epgs_group"

    class Origin(Enum):
        DEVICE = "device"

    class IPType(Enum):
        IPV4 = "IPV4"
        IPV6 = "IPV6"
        OTHER = "OTHER"

    class Vendor(Enum):
        CISCO = "Cisco"
        FORTINET = "Fortinet"
        NETSCREEN = "Netscreen"
        CHECKPOINT = "Checkpoint"
        PALO_ALTO = "PaloAltoNetworks"
        NEW_F5 = "NewF5"
        F5 = "f5"
        MCAFEE = "Mcafee"
        STONESOFT = "Stonesoft"
        BLUECOAT = "bluecoat"
        GENERIC = "Generic"
        LINUX = "linux"
        VMWARE = "VMware"
        AMAZON = "Amazon"
        OPENSTACK = "OpenStack"
        AZURE = "Azure"

    class Prop(Enum):
        GLOBAL = "global"
        IMPLICIT = "implicit"
        SHARED = "shared"

    xsi_type: Optional[XsiType] = prop(None, key=Jsonable.Prop.XSI_TYPE.value)
    uid: str = prop("")
    name: Optional[str] = prop(None)
    display_name: Optional[str] = prop(None)
    class_name: Optional[ClassName] = prop(None)
    origin: Optional[Origin] = prop(None)
    is_global: bool = prop(False, key=Prop.GLOBAL.value)
    is_implicit: bool = prop(False, key=Prop.IMPLICIT.value)
    is_shared: bool = prop(False, key=Prop.SHARED.value)
    comment: Optional[str] = prop(None)
    id: Union[str, int] = prop()
    version_id: int = prop(0)
    device_type: Optional[Vendor] = prop(None)  # This seems to be the vendor field.
    ip_type: Optional[IPType] = prop(None)
    referenced: str = prop("")
    installable_target: bool = prop(False)

    @property
    def implicit(self):
        return self.is_implicit

    @implicit.setter
    def implicit(self, v):
        self.is_implicit = v

    @property
    def shared(self):
        return self.is_shared

    @shared.setter
    def shared(self, v):
        self.is_shared = v


@propify
class AnyNetworkObject(NetworkObject):
    xsi_type: NetworkObject.XsiType = prop(NetworkObject.XsiType.ANY_NETWORK_OBJECT)


@propify
class HostNetworkObject(NetworkObject):
    xsi_type: NetworkObject.XsiType = prop(NetworkObject.XsiType.HOST_NETWORK_OBJECT)
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    netmask: Optional[IPAddress] = prop(
        converter=optional(IPAddress),
        jsonify=stringify_optional_obj,
        factory=lambda: IPAddress("255.255.255.255"),
    )


@propify
class InterfaceIP(Jsonable):
    ip: Optional[IPAddress] = prop(None)
    subnet_mask: Optional[IPAddress] = prop(None)


@propify
class Interface(Jsonable):
    name: Optional[str] = prop(None)
    anti_spoof_disabled: Optional[bool] = prop(None)
    interface_ips: List[InterfaceIP] = prop(factory=list, flatify="interface_ip")


@propify
class HostNetworkObjectWithInterfaces(HostNetworkObject):
    interfaces: List[Interface] = prop(factory=list, key="interface")


@propify
class SubnetObject(NetworkObject):
    ip: Optional[str] = prop(None)
    subnet_mask: Optional[str] = prop(None)


@propify
class RangeObject(NetworkObject):
    min_ip: Optional[str] = prop(None)
    max_ip: Optional[str] = prop(None)


@propify
class HostObject(HostNetworkObject):
    # NOT NEEDED?
    pass


def classify_object_type(network_object):
    if "@xsi.type" not in network_object:
        return NetworkObject.kwargify(network_object)

    if network_object["@xsi.type"] in (
        NetworkObject.XsiType.HOST_NETWORK_OBJECT.value,
        NetworkObject.XsiType.HOST_NETWORK_OBJECT_NO_NAMESPACE.value,
    ):
        return HostObject.kwargify(network_object)
    elif network_object["@xsi.type"] == NetworkObject.XsiType.ANY_NETWORK_OBJECT.value:
        return AnyNetworkObject.kwargify(network_object)
    elif network_object["@xsi.type"] == (
        NetworkObject.XsiType.HOST_NETWORK_OBJECT_WITH_INTERFACES.value
    ):
        return HostNetworkObjectWithInterfaces.kwargify(network_object)
    elif (
        network_object["@xsi.type"] == NetworkObject.XsiType.SUBNET_NETWORK_OBJECT.value
    ):
        return SubnetObject.kwargify(network_object)
    elif (
        network_object["@xsi.type"] == NetworkObject.XsiType.RANGE_NETWORK_OBJECT.value
    ):
        return RangeObject.kwargify(network_object)
    elif (
        network_object["@xsi.type"] == NetworkObject.XsiType.NETWORK_OBJECT_GROUP.value
    ):
        return GroupObject.kwargify(network_object)
    else:
        return NetworkObject.kwargify(network_object)


@propify
class GroupObject(NetworkObject):
    class Prop(Enum):
        MEMBER = "member"

    members: List[Union[HostObject, SubnetObject, RangeObject]] = prop(
        factory=list, repr=False, key=Prop.MEMBER.value, kwargify=classify_object_type
    )



from pytos2.api import BaseAPI, Pager
import pytest

PAGER_TEST_LIST = [{"id": idx} for idx in range(10000)]


class _TestResponse:
    def __init__(self, data):
        self.data = data

    def json(self):
        return self.data

    @property
    def ok(self):
        return True


class _TestSession:
    def __init__(self, emulate_total=True):
        self.emulate_total = emulate_total

    def get(self, *args, params=None, **kwargs):
        params = params or {}

        tlist = PAGER_TEST_LIST[
            params.get("start", 0) : params.get("start", 0) + params.get("count", 0)
        ]
        data = {"data": {"count": len(tlist), "data": tlist}}

        if params.get("get_total", None) == "true" and self.emulate_total:
            data["data"]["total"] = len(PAGER_TEST_LIST)

        return _TestResponse(data)


class _TestAPI(BaseAPI):
    def __init__(self, emulate_total=True):
        super().__init__(_TestSession(emulate_total=emulate_total))


class _TestPager(Pager):
    def __init__(
        self,
        api_node="data.data",
        page_size=2000,
        start=None,
        stop=None,
        step=None,
        is_slice=False,
        emulate_total=True,
    ):
        self.emulate_total = emulate_total

        super().__init__(
            _TestAPI(emulate_total=emulate_total),
            None,
            api_node,
            None,
            None,
            page_size=page_size,
            start=start,
            stop=stop,
            step=step,
            is_slice=is_slice,
        )

    def _new_pager(self, start=None, stop=None, step=None, is_slice=False):
        return _TestPager(
            page_size=self.page_size,
            start=start,
            stop=stop,
            step=step,
            is_slice=is_slice,
            emulate_total=self.emulate_total,
        )


class TestPager:
    def test_basic_pager(self):
        basic = _TestPager()

        assert basic[0]["id"] == 0
        assert basic[2001]["id"] == 2001

        assert isinstance(basic, Pager)

        assert len(basic) == 10000

        # Cover the emulated total instance.
        assert basic.fetch_all()[-1]["id"] == 9999

    def test_one_slice(self):
        basic = _TestPager()
        sliced = basic[5:10]

        assert isinstance(sliced, Pager)
        assert isinstance(sliced, _TestPager)

        assert sliced[0]["id"] == 5
        assert sliced[1]["id"] == 6
        assert sliced[2]["id"] == 7
        assert sliced[3]["id"] == 8
        assert sliced[4]["id"] == 9

        with pytest.raises(IndexError):
            sliced[5]

        assert len(sliced) == 5
        sliced_repr = repr(sliced)
        assert sliced is iter(sliced)

        assert "Pager" in sliced_repr
        assert "5 items" in sliced_repr

    def test_slice_of_slice(self):
        basic = _TestPager()
        sliced = basic[5:10]
        second = sliced[1:3]
        assert isinstance(second, list)
        assert len(second) == 2
        assert second[0]["id"] == 6

    def test_slice_with_step(self):
        basic = _TestPager()
        sliced = basic[5:10:2]
        assert sliced[0]["id"] == 5
        assert sliced[1]["id"] == 7
        assert len(sliced) == 2

        assert isinstance(sliced, Pager)

        sliced = basic[10::2]
        assert sliced[0]["id"] == 10
        assert sliced[-1]["id"] == 9998
        assert len(sliced) == 4995

    def test_slice_with_only_start(self):
        basic = _TestPager()
        sliced = basic[10:]
        assert sliced[0]["id"] == 10
        assert sliced[-1]["id"] == 9999
        assert len(sliced) == 9990

    def test_slice_with_only_stop(self):
        basic = _TestPager()
        sliced = basic[:10]
        assert sliced[0]["id"] == 0
        assert sliced[-1]["id"] == 9
        assert len(sliced) == 10

    def test_slice_with_only_step(self):
        basic = _TestPager()
        sliced = basic[::2]
        assert sliced[0]["id"] == 0
        assert sliced[1]["id"] == 2
        assert sliced[-1]["id"] == 9998
        assert len(sliced) == 5000

    def test_next(self):
        basic = _TestPager()
        first = next(basic)
        assert first["id"] == 0

        last_id = None
        for item in basic:
            last_id = item["id"]

        assert last_id == 9999

    def test_eq(self):
        basic = _TestPager()
        basic2 = _TestPager()

        sliced = basic[5:10]

        assert basic == basic2
        assert sliced == basic2[5:10]
        assert basic != sliced

        assert basic != 0

    def test_invalid_slices(self):
        basic = _TestPager()

        with pytest.raises(IndexError):
            basic[-5:10:2]

        with pytest.raises(IndexError):
            basic[5:-10:2]

        with pytest.raises(IndexError):
            basic[5:10:-2]

    def test_invalid_indices(self):
        basic = _TestPager()

        with pytest.raises(IndexError):
            basic[5:10][10]

        with pytest.raises(IndexError):
            basic[10500]

    def test_no_total_from_api(self):
        basic = _TestPager(emulate_total=False)

        assert len(basic) == 2000
        assert basic.fetch_all()[-1]["id"] == 9999

        assert len(basic) == 10000



from datetime import date, datetime
from typing import Union, Optional, List

from requests import Response
from requests.exceptions import HTTPError

# avoid circular imports
import pytos2
from pytos2.api import ApiError
from pytos2.models import DateFilterType, coerce_date_filter_type  # noqa
from .api import SaAPI
from pytos2.utils import NoInstance, get_api_node
from .application_identities import ApplicationIdentity
from .applications import Application, ApplicationConnection
from .history import HistoryRecord
from pytos2.securechange.user import SCWUser
from pytos2.models import ObjectReference


class Sa:
    default: Union["Sa", NoInstance] = NoInstance(
        "Sa.default",
        "No Sa instance has been initialized yet, initialize with `Sa(*args, **kwargs)`",
    )

    def __init__(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        default=True,
    ):
        self.api: SaAPI = SaAPI(hostname, username, password)
        if default:
            Sa.default = self

        self._application_identities: List[ApplicationIdentity] = []

    @property
    def application_identities(self) -> List[ApplicationIdentity]:
        if not self._application_identities:
            res = self._get_application_identities()
            self._application_identities = res
        return self._application_identities

    def _get_application_identities(self, cache=True) -> List[ApplicationIdentity]:
        res = self.api.get_application_identities()
        data = self.api.handle_json(res, "application_identities")
        return [
            ApplicationIdentity.kwargify(a)
            for a in get_api_node(
                data, "application_identities.application_identity", listify=True
            )
        ]

    def get_applications(self) -> List[Application]:
        """
        This function returns a list of all applications in SecureApp.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications

        Usage:
            apps = sa.get_applications()
        """

        res = self.api.get_applications()
        data = self.api.handle_json(res, "get_applications")
        return [
            Application.kwargify(a)
            for a in get_api_node(data, "applications.application", listify=True)
        ]

    def get_application(self, application_id: str) -> Application:
        """
        This function returns a single application by its ID.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60

        Usage:
            app = sa.get_application(60)
        """

        res = self.api.get_application(application_id)
        data = self.api.handle_json(res, "get_application")
        application = get_api_node(data, "application")
        if isinstance(application, list) and application:
            application = application[0]

        return Application.kwargify(application)

    def add_application(
        self,
        name: Optional[str] = None,
        comment: Optional[str] = None,
        owner: Union[None, SCWUser, int, str] = None,
        editors: Optional[List[Union[SCWUser, int]]] = None,
        viewers: Optional[List[Union[SCWUser, int]]] = None,
        customer: Optional[
            int
        ] = None,  # TODO: Add customer to this once it's implemented.
    ):
        """
        This function creates a new application.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/applications

        Usage:
            Example 1:
            app = sa.add_application(name="My App", comment="This is my app")

            Example 2:
            app = sa.add_application(
                name="My App",
                owner=100,
                editors=[101, 102],
                viewers=[103, 104]
            )
        """

        from pytos2.securechange import Scw

        if isinstance(owner, str):
            owner = Scw.default.get_user(owner)
        if owner:
            owner = owner if isinstance(owner, int) else owner.id

        editors = editors or []
        viewers = viewers or []

        editors = [e if isinstance(e, int) else e.id for e in editors]
        viewers = [v if isinstance(v, int) else v.id for v in viewers]

        app = Application(
            name=name,
            comment=comment,
            owner=ObjectReference(id=owner) if owner else None,
            editors=[ObjectReference(id=e.id) for e in editors],
            viewers=[ObjectReference(id=v.id) for v in viewers],
            customer=ObjectReference(id=customer) if customer else None,
        )

        res = self.api.add_application(app._json)
        res = self.api.handle_response(res, "add_application")
        if res.status_code == 201:
            loc = res.headers.get("Location", "")
            new_app_id = loc.split("/")[-1]
            return self.get_application(new_app_id)
        else:
            # This should never happen, but just in case...
            raise ApiError(
                f"Failed to add application from status code: {res.status_code} and response: {res.text}"
            )  # noqa

    def update_application(
        self,
        application_id: Union[int, Application],
        name: Optional[str] = None,
        comment: Optional[str] = None,
        owner: Union[None, SCWUser, int] = None,
        editors: Optional[List[Union[SCWUser, int]]] = None,
        viewers: Optional[List[Union[SCWUser, int]]] = None,
    ):
        """
        This function updates an existing application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60

        Usage:
            Example 1:
            sa.update_application(60, name="My App", comment="This is my app")

            Example 2:
            sa.update_application(
                60,
                name="My App",
                owner=100,
                editors=[101, 102],
                viewers=[103, 104]
            )
        """
        if owner:
            owner = owner if isinstance(owner, int) else owner.id

        editors = (
            [e if isinstance(e, int) else e.id for e in editors] if editors else None
        )
        viewers = (
            [v if isinstance(v, int) else v.id for v in viewers] if viewers else None
        )

        if isinstance(application_id, Application):
            app = application_id
            application_id = app.id
        else:
            app = Application(
                name=name,
                comment=comment,
                owner=ObjectReference(id=owner) if owner else None,
                editors=(
                    [ObjectReference(id=e.id) for e in editors] if editors else None
                ),
                viewers=(
                    [ObjectReference(id=v.id) for v in viewers] if viewers else None
                ),
            )

        res = self.api.update_application(application_id, app._json)
        res = self.api.handle_response(res, "update_application")

    def delete_application(self, application_id: int):
        """
        This function deletes an application by its ID.

        Method: DELETE
        URL: /securechangeworkflow/api/secureapp/repository/applications/60

        Usage:
            sa.delete_application(60)
        """

        res = self.api.delete_application(application_id)
        self.api.handle_response(res, "delete_application")

    def bulk_update_applications(self, apps: List[Application]):
        res = self.api.update_applications([a._json for a in apps])
        self.api.handle_response(res, "bulk_update_applications")

    def get_application_connections(
        self, application_id: int
    ) -> List[ApplicationConnection]:
        """
        This function returns a list of all connections for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections

        Usage:
            connections = sa.get_application_connections(60)
        """

        res = self.api.get_application_connections(application_id)
        data = self.api.handle_json(res, "get_application_connections")
        connections = get_api_node(data, "connections.connection", listify=True)
        return [ApplicationConnection.kwargify(c) for c in connections]

    def get_application_connection(
        self, application_id: int, connection_id: int
    ) -> ApplicationConnection:
        """
        This function returns a single connection for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections/100

        Usage:
            connection = sa.get_application_connection(60, 100)
        """

        res = self.api.get_application_connection(application_id, connection_id)
        data = self.api.handle_json(res, "get_application_connection")
        conn = get_api_node(data, "connection")
        return ApplicationConnection.kwargify(conn)

    def get_application_history(
        self,
        application_id: int,
        start_date: DateFilterType = None,
        end_date: DateFilterType = None,
        count: Optional[int] = None,
        start: Optional[int] = None,
        type: Optional[str] = None,
        user: Optional[str] = None,
        show_members: Optional[bool] = None,
    ):
        """
        This function returns the history of a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/history

        Usage:
            history = sa.get_application_history(60)
        """

        start_date = coerce_date_filter_type(start_date)
        end_date = coerce_date_filter_type(end_date)

        res = self.api.get_application_history(
            application_id,
            start_date=start_date,
            end_date=end_date,
            count=count,
            start=start,
            type=type,
            user=user,
            show_members=show_members,
        )
        data = self.api.handle_json(res, "get_application_history")
        history = get_api_node(data, "history_records.history_record", listify=True)
        return [HistoryRecord.kwargify(h) for h in history]



from enum import Enum

from typing import Optional, List, Union
from netaddr import IPAddress
from attr.converters import optional

from pytos2.utils import propify, prop, stringify_optional_obj
from pytos2.models import Jsonable

from pytos2.securechange.rule import SlimRule
from pytos2.securechange.designer import DesignerResults
from pytos2.securechange.designer_verifier_common import NetworkObject

from pytos2.securetrack import St
from pytos2.securetrack.service_object import Service as STService

from pytos2.securetrack.rule import SecurityRule
from pytos2.securetrack.network_object import (
    NetworkObjectXsiType as STNetworkObjectXsiType,
    NetworkObject as STNetworkObject,
)

from .binding import Binding
from . import Designer, Field, FieldXsiType


def _mv_key(j, src, dest):
    if src not in j:
        return

    j[dest] = j[src]
    del j[src]


@propify
class ModificationNetworkObject(Jsonable):
    class XsiType(Enum):
        NETWORK_OBJECT = "ns_sc_policy:network_object"
        SUBNET_NETWORK_OBJECT = "ns_sc_policy:subnet_network_object"
        HOST_NETWORK_OBJECT = "ns_sc_policy:host_network_object"
        RANGE_NETWORK_OBJECT = "ns_sc_policy:range_network_object"
        NETWORK_OBJECT_GROUP = "ns_sc_policy:network_object_group"
        HOST_NETWORK_OBJECT_WITH_INTERFACES = (
            "ns_sc_policy:host_network_object_with_interfaces"
        )

    xsi_type: XsiType = prop(XsiType.NETWORK_OBJECT)
    id: Optional[int] = prop(None)
    comment: Optional[str] = prop(None)
    name: Optional[str] = prop(None)

    def from_securetrack(obj):
        j = dict(obj.data or obj._json)
        if obj.xsi_type == STNetworkObjectXsiType.RANGE_NETWORK_OBJECT:
            _mv_key(j, "first_ip", "min_ip")
            _mv_key(j, "last_ip", "max_ip")

        cls = {
            STNetworkObjectXsiType.HOST_NETWORK_OBJECT: ModificationHostNetworkObject,
            STNetworkObjectXsiType.HOST_NETWORK_OBJECT_WITH_INTERFACES: ModificationHostNetworkObjectWithInterfaces,
            STNetworkObjectXsiType.NETWORK_OBJECT_GROUP: ModificationNetworkObjectGroup,
            STNetworkObjectXsiType.RANGE_NETWORK_OBJECT: ModificationRangeNetworkObject,
            STNetworkObjectXsiType.SUBNET_NETWORK_OBJECT: ModificationSubnetNetworkObject,
        }.get(obj.xsi_type, ModificationNetworkObject)
        return cls.kwargify(j)


@propify
class ModificationHostNetworkObject(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.HOST_NETWORK_OBJECT
    )
    netmask: Optional[IPAddress] = prop(
        IPAddress("255.255.255.255"),
        converter=optional(IPAddress),
        jsonify=stringify_optional_obj,
    )
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )


@propify
class ModificationSubnetNetworkObject(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.SUBNET_NETWORK_OBJECT
    )
    netmask: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )


@propify
class ModificationRangeNetworkObject(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.RANGE_NETWORK_OBJECT
    )
    min_ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    max_ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )


@propify
class ModificationNetworkObjectGroup(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.NETWORK_OBJECT_GROUP
    )


@propify
class ModificationHostNetworkObjectWithInterfaces(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.HOST_NETWORK_OBJECT_WITH_INTERFACES
    )
    netmask: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )


@propify
class RangeDTO(Jsonable):
    class Prop(Enum):
        FROM = "from"

    to: int = prop()
    from_: int = prop(key=Prop.FROM.value)


@propify
class ModificationServiceObject(Jsonable):
    class XsiType(Enum):
        SERVICE_OBJECT = "ns_sc_policy:service_object"
        ICMP_SERVICE = "ns_sc_policy:icmp_service"
        IP_SERVICE = "ns_sc_policy:ip_service"
        SERVICE_GROUP = "ns_sc_policy:service_group"
        TRANSPORT_SERVICE = "ns_sc_policy:transport_service"

    xsi_type: XsiType = prop(XsiType.SERVICE_OBJECT, key=Jsonable.Prop.XSI_TYPE.value)

    id: Optional[int] = prop(None)
    comment: Optional[str] = prop(None)
    name: Optional[str] = prop(None)

    def from_securetrack(obj):
        from pytos2.securetrack.service_object import (
            TCPServiceObject,
            UDPServiceObject,
            ICMPServiceObject,
            ServiceGroup,
            OtherIPServiceObject,
        )

        if isinstance(obj, (TCPServiceObject, UDPServiceObject)):
            protos = {6: "TCP", 17: "UDP"}

            proto = protos.get(obj.protocol, obj.protocol)

            return ModificationTransportService(
                comment=obj.comment,
                name=obj.name,
                protocol=proto,
                port=RangeDTO(from_=obj.min_port, to=obj.max_port),
            )
        elif isinstance(obj, (ICMPServiceObject)):
            return ModificationIcmpService(
                comment=obj.comment,
                name=obj.name,
                type=RangeDTO(from_=obj.min_port, to=obj.max_port),
            )
        elif isinstance(obj, (ServiceGroup)):
            return ModificationServiceGroup(comment=obj.comment, name=obj.name)
        elif isinstance(obj, (OtherIPServiceObject)):
            return ModificationIPService(comment=obj.comment, name=obj.name)
        else:
            return ModificationServiceObject(comment=obj.comment, name=obj.name)


@propify
class ModificationIcmpService(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.ICMP_SERVICE, key=Jsonable.Prop.XSI_TYPE.value
    )

    type: Optional[RangeDTO] = prop(None)


@propify
class ModificationIPService(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.IP_SERVICE, key=Jsonable.Prop.XSI_TYPE.value
    )

    protocol: Optional[RangeDTO] = prop(None)


@propify
class ModificationServiceGroup(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.SERVICE_GROUP,
        key=Jsonable.Prop.XSI_TYPE.value,
    )


@propify
class ModificationTransportService(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.TRANSPORT_SERVICE,
        key=Jsonable.Prop.XSI_TYPE.value,
    )

    protocol: Optional[str] = prop(None)
    port: Optional[RangeDTO] = prop(None)


@propify
class RuleOperationBinding(Jsonable):
    binding_uid: str = prop()
    binding: Optional[Binding] = prop(None)
    rules: List[SlimRule] = prop(factory=list, flatify="rule")


@propify
class RuleOperationDevice(Jsonable):
    revision_id: Optional[int] = prop(None)
    management_id: int = prop()

    management_name: str = prop("")
    management_ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )

    revision_number: Optional[int] = prop(None)
    number_of_rules: Optional[int] = prop(None)

    bindings: List[RuleOperationBinding] = prop(factory=list, flatify="binding")


@propify
class RuleOperation(Field):
    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_RECERTIFICATION, key=Jsonable.Prop.XSI_TYPE.value
    )

    devices: List[RuleOperationDevice] = prop(factory=list, flatify="device")
    _ticketed_rules: Optional[Union[List[SecurityRule], None]] = None

    def get_rules(self):
        rules = []

        for device in self.devices:
            for binding in device.bindings:
                rules += binding.rules

        return rules

    def is_rule_added(self, rule: SecurityRule):
        rules = self.get_rules()
        for slim in rules:
            if slim.uid == rule.uid:
                return True

        return False

    def _add_devices(self, rules: list):
        for rule in rules:
            device = RuleOperationDevice(
                management_id=rule.device.id,
                management_name=rule.device.name,
                revision_id=rule.device.latest_revision,
            )
            if device.management_id not in [d.management_id for d in self.devices]:
                self.devices.append(device)

    def is_invalid_rule(self, rule: SecurityRule):
        def skip_invalid_rule(msg):
            # LOGGER.info(msg)
            return True

        rule_uid = rule.uid

        if self._ticketed_rules is None:
            self._ticketed_rules = St.default.get_rules_on_open_tickets()

        for tr in self._ticketed_rules:
            if rule.id == tr.id:
                return skip_invalid_rule(
                    "Rule {} {} is already in an active ticket, skipping.".format(
                        rule.id, rule.uid
                    )
                )

        if rule.rule_location is not None:
            if rule.rule_location.lower() in ["predefined", "shared"]:
                return skip_invalid_rule(
                    "Rule {} {} has location of {}, skipping.".format(
                        rule.id, rule_uid, rule.rule_location
                    )
                )
        if rule.rule_type is not None:
            if rule.rule_type.value.lower() in ["interzone", "intrazone"]:
                return skip_invalid_rule(
                    "Rule {} {} has type of {}, skipping.".format(
                        rule.id, rule_uid, rule.rule_type
                    )
                )

        if rule.device.model.value.lower() in ["fortimanager", "fmg_adom"]:
            if rule.global_location and rule.global_location.value.lower() in [
                "before",
                "after",
            ]:
                return skip_invalid_rule(
                    "Rule {} {} is global, skipping.".format(rule.id, rule_uid)
                )

        if (
            rule.device.vendor.value.lower() == "paloaltonetworks"
            and rule.global_location
        ):
            if rule.rule_location is not None:
                if rule.rule_location != rule.device.context_name:
                    return skip_invalid_rule(
                        "Rule {} {} is not located on the current device context, skipping.".format(
                            rule.id, rule_uid
                        )
                    )
            elif rule.device.parent is not None:
                if rule.global_location.value.lower() in ["before", "after"]:
                    return skip_invalid_rule(
                        "Rule {} {} is global, skipping.".format(rule.id, rule_uid)
                    )

        if self.xsi_type == FieldXsiType.RULE_RECERTIFICATION:
            if rule.is_disabled:
                return skip_invalid_rule(
                    "Rule {} {} - rule disabled, skipping.".format(rule.id, rule_uid)
                )
            if (
                rule.documentation is not None
                and rule.documentation.certification_status is not None
            ):
                if (
                    rule.documentation.certification_status.value.lower()
                    == "decertified"
                ):
                    return skip_invalid_rule(
                        "Rule {} {} - rule decertified, skipping.".format(
                            rule.id, rule_uid
                        )
                    )
        return False

    def add_rule(
        self,
        rule: SecurityRule,
        skip_invalid_rules: Optional[bool] = False,
    ):
        self.add_rules([rule], skip_invalid_rules)

    def add_rules(
        self,
        rules: List[SecurityRule],
        skip_invalid_rules: Optional[bool] = False,
    ):
        """
        Usage Notes:

        If skip_invalid_rules = True, there will be no checks to see if a rule can be added to a ticket
        and the ticket save/post may cause an error.

        If skip_invalid_rules = False, some attempt will be made to discard rules which cannot be added to tickets.
        """

        self._add_devices(rules)

        for device in self.devices:
            for rule in rules:
                if (
                    device.management_id == rule.device.id
                    and device.revision_id == rule.device.latest_revision
                ):
                    if skip_invalid_rules and self.is_invalid_rule(rule):
                        continue

                    for binding in device.bindings:
                        if binding.binding_uid == rule.bindings[0].uid:
                            break
                    else:
                        binding = RuleOperationBinding(binding_uid=rule.bindings[0].uid)
                        device.bindings.append(binding)

                    slim_rule = SlimRule.kwargify({"uid": rule.uid})
                    binding.rules.append(slim_rule)


@propify
class ObjectStatus(Jsonable):
    class XsiType(Enum):
        EXISTING = "existing"
        NEW = "new"

    xsi_type: XsiType = prop(XsiType.EXISTING, key=Jsonable.Prop.XSI_TYPE.value)
    st_uid: Optional[str] = prop(None)


@propify
class DeviceNetworkObject(Jsonable):
    network_object: List[NetworkObject] = prop(factory=list)
    device_id: Optional[int] = prop(None)
    status: Optional[ObjectStatus] = prop(None)


@propify
class DeviceServiceObject(Jsonable):
    service_object: List[ModificationServiceObject] = prop(factory=list)
    uid: Optional[str] = prop(None)
    device_id: Optional[int] = prop(None)
    status: Optional[ObjectStatus] = prop(None)
    id: Optional[int] = prop(None)


class CellAction(Enum):
    REMOVE = "REMOVE"
    ADD = "ADD"


@propify
class NetworkObjectCellModification(Jsonable):
    device_network_object: DeviceNetworkObject = prop()
    action: Optional[CellAction] = prop(None)
    id: Optional[int] = prop(None)


@propify
class NetworkObjectCellModifications(Jsonable):
    id: Optional[int] = prop(None)
    network_object_cell_modifications: List[NetworkObjectCellModification] = prop(
        factory=list, flatify="network_object_cell_modification"
    )


@propify
class ServiceObjectCellModification(Jsonable):
    device_service_object: DeviceServiceObject = prop()
    action: Optional[CellAction] = prop(None)
    id: Optional[int] = prop(None)


@propify
class ServiceObjectCellModifications(Jsonable):
    id: Optional[int] = prop(None)
    service_object_cell_modifications: List[ServiceObjectCellModification] = prop(
        factory=list, flatify="service_object_cell_modification"
    )


@propify
class RuleKey(Jsonable):
    device_id: Optional[int] = prop(None)
    binding_uid: Optional[str] = prop(None)
    rule_uid: Optional[str] = prop(None)

    def __eq__(a, b):
        return all(
            [
                a.device_id == b.device_id,
                a.binding_uid == b.binding_uid,
                a.rule_uid == b.rule_uid,
            ]
        )


@propify
class RuleModification(Jsonable):
    class XsiType(Enum):
        RULE_MODIFICATION = "rule_modification"
        MODIFY_RULE_MODIFICATION = "modify_rule_modification"

    rule_key: RuleKey = prop()
    id: Optional[int] = prop(None)
    xsi_type: XsiType = prop(
        XsiType.RULE_MODIFICATION, key=Jsonable.Prop.XSI_TYPE.value
    )

    _rule: SecurityRule = prop(None, jsonify=False)

    @property
    def rule(self):
        if self._rule:
            return self._rule

        rules = St.default.get_rules(
            device=self.rule_key.device_id, rule_uid=self.rule_key.rule_uid
        )
        rule = rules[0]
        self._rule = rule
        return self._rule


@propify
class ModifyRuleModification(RuleModification):
    xsi_type: RuleModification.XsiType = prop(
        RuleModification.XsiType.MODIFY_RULE_MODIFICATION,
        key=Jsonable.Prop.XSI_TYPE.value,
    )
    source_modifications: Optional[NetworkObjectCellModifications] = prop(None)
    destination_modifications: Optional[NetworkObjectCellModifications] = prop(None)
    service_modifications: Optional[ServiceObjectCellModifications] = prop(None)

    def _add_service_obj_mod(self, modlist, objlist, obj, device_id, cell_action):
        if isinstance(obj, str):
            obj = St.default.get_service(obj, device_id)
            _uid = obj.uid

            obj = ModificationServiceObject.from_securetrack(obj)
        elif isinstance(obj, STService):
            _uid = obj.uid

            obj = ModificationServiceObject.from_securetrack(obj)
        else:
            _uid = None

        if not obj:
            return False

        obj_status = ObjectStatus(
            xsi_type=ObjectStatus.XsiType.EXISTING if _uid else ObjectStatus.XsiType.NEW
        )
        if _uid:
            obj_status.st_uid = _uid

        svc_mod = ServiceObjectCellModification(
            action=cell_action,
            device_service_object=DeviceServiceObject(
                device_id=device_id, status=obj_status
            ),
        )

        found = False
        for ref in objlist:
            if obj_status.st_uid and ref.uid == obj_status.st_uid:
                found = True
                break

        if (cell_action == CellAction.REMOVE and found) or (
            cell_action == CellAction.ADD and not found
        ):
            svc_mod.device_service_object.service_object.append(obj)
            modlist.append(svc_mod)
            return True
        else:
            return False

    def _add_net_obj_mod(self, modlist, objlist, obj, device_id, cell_action):
        if isinstance(obj, str):
            obj = St.default.get_network_object(obj, device_id)
            _uid = obj.uid
            obj = ModificationNetworkObject.from_securetrack(obj)
        elif isinstance(obj, STNetworkObject):
            _uid = obj.uid
            obj = ModificationNetworkObject.from_securetrack(obj)
        else:
            _uid = None

        if not obj:
            return False

        obj_status = ObjectStatus(
            xsi_type=ObjectStatus.XsiType.EXISTING if _uid else ObjectStatus.XsiType.NEW
        )
        if _uid:
            obj_status.st_uid = _uid

        net_mod = NetworkObjectCellModification(
            action=cell_action,
            device_network_object=DeviceNetworkObject(
                device_id=device_id, status=obj_status
            ),
        )

        found = False
        for ref in objlist:
            if obj_status.st_uid and ref.uid == obj_status.st_uid:
                found = True
                break

        if (cell_action == CellAction.REMOVE and found) or (
            cell_action == CellAction.ADD and not found
        ):
            net_mod.device_network_object.network_object.append(obj)
            modlist.append(net_mod)
            return True
        else:
            return False

    def add_source(self, obj: Union[str, NetworkObject], device_id):
        if not self.source_modifications:
            self.source_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.source_modifications.network_object_cell_modifications,
            self.rule.src_networks,
            obj,
            device_id,
            CellAction.ADD,
        )

    def remove_source(self, obj: Union[str, NetworkObject], device_id):
        if not self.source_modifications:
            self.source_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.source_modifications.network_object_cell_modifications,
            self.rule.src_networks,
            obj,
            device_id,
            CellAction.REMOVE,
        )

    def add_destination(self, obj: Union[str, NetworkObject], device_id):
        if not self.destination_modifications:
            self.destination_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.destination_modifications.network_object_cell_modifications,
            self.rule.dest_networks,
            obj,
            device_id,
            CellAction.ADD,
        )

    def remove_destination(self, obj: Union[str, NetworkObject], device_id):
        if not self.destination_modifications:
            self.destination_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.destination_modifications.network_object_cell_modifications,
            self.rule.dest_networks,
            obj,
            device_id,
            CellAction.REMOVE,
        )

    def add_service(self, obj, device_id):
        if not self.service_modifications:
            self.service_modifications = ServiceObjectCellModifications()

        return self._add_service_obj_mod(
            self.service_modifications.service_object_cell_modifications,
            self.rule.src_services + self.rule.dest_services,
            obj,
            device_id,
            CellAction.ADD,
        )

    def remove_service(self, obj, device_id):
        if not self.service_modifications:
            self.service_modifications = ServiceObjectCellModifications()

        return self._add_service_obj_mod(
            self.service_modifications.service_object_cell_modifications,
            self.rule.src_services + self.rule.dest_services,
            obj,
            device_id,
            CellAction.REMOVE,
        )


def classify_rule_modification(obj):
    return (
        {"modify_rule_modification": ModifyRuleModification}
        .get(obj.get(Jsonable.Prop.XSI_TYPE.value))
        .kwargify(obj)
    )


@propify
class RuleModificationField(RuleOperation):
    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_MODIFICATION_FIELD, key=Jsonable.Prop.XSI_TYPE.value
    )

    rule_modifications: List[RuleModification] = prop(
        factory=list, flatify="rule_modification", kwargify=classify_rule_modification
    )

    def _get_rule_mod_for_rule(self, rule: SecurityRule):
        rule_key = RuleKey(
            device_id=rule.device.id,
            binding_uid=rule.bindings[0].uid,
            rule_uid=rule.uid,
        )

        for mod in self.rule_modifications:
            if mod.rule_key == rule_key:
                return mod

        mod = ModifyRuleModification(rule_key=rule_key)
        mod._rule = rule

        self.rule_modifications.append(mod)

        return mod

    def add_rule_modification(self, rule: SecurityRule) -> RuleModification:
        if not self.is_rule_added(rule):
            self.add_rule(rule)

        mod = self._get_rule_mod_for_rule(rule)
        return mod

    def add_source_object(self, rule: SecurityRule, obj: Union[str, NetworkObject]):
        return self.add_rule_modification(rule).add_source(obj, rule.device.id)

    def remove_source_object(self, rule: SecurityRule, obj: Union[str, NetworkObject]):
        return self.add_rule_modification(rule).remove_source(obj, rule.device.id)

    def add_destination_object(
        self, rule: SecurityRule, obj: Union[str, NetworkObject]
    ):
        return self.add_rule_modification(rule).add_destination(obj, rule.device.id)

    def remove_destination_object(
        self, rule: SecurityRule, obj: Union[str, NetworkObject]
    ):
        return self.add_rule_modification(rule).remove_destination(obj, rule.device.id)

    def add_service(self, rule: SecurityRule, obj: ModificationServiceObject):
        return self.add_rule_modification(rule).add_service(obj, rule.device.id)

    def remove_service(self, rule: SecurityRule, obj: ModificationServiceObject):
        return self.add_rule_modification(rule).remove_service(obj, rule.device.id)


@propify
class RuleRecertification(RuleOperation):
    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_RECERTIFICATION, key=Jsonable.Prop.XSI_TYPE.value
    )


@propify
class RuleDecommission(RuleOperation):
    class Action(Enum):
        DISABLE = "disable"
        REMOVE = "remove"

    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_DECOMMISSION, key=Jsonable.Prop.XSI_TYPE.value
    )

    designer_result: Optional[Designer] = prop(None)
    verifier_result: Optional[dict] = prop(None)

    action: Optional[Action] = prop(None)



import json
import pytest
import responses


class TestWorkflows:
    @responses.activate
    def test_get_domains(self, domains_mock, scw):
        domains = scw.get_domains()
        assert domains[8].id == "9"
        assert domains[8].name == "1Toronto BCKP"
        assert domains[8].description == ""

    @responses.activate
    def test_get_domain(self, domains_mock, scw):
        domain = scw.get_domain(2)
        assert domain.id == "2"
        assert domain.name == "Toronto"
        assert domain.description == "Test"



import pytest
import json
import responses

from pytos2.securechange.user import (
    NewLocalUser,
    LDAPConfiguration,
    LDAPParty,
    LDAPSimulateLogin,
    LDAPUserSimulateLogin,
)


class TestUsers:
    @responses.activate
    def test_add_user_group(self, user_or_group_mock, user_mock, scw):
        res = scw.add_user_group("Group1", "Test Group 1", "group1@contoso.com")
        assert res.email == "contoso@contoso.com"

    @responses.activate
    def test_update_user_group(self, user_or_group_mock, scw):
        res = scw.update_user_group(54, ["70", "96"], ["85", "49"])
        assert res is None

        res = scw.update_user_group(54, ["70", "96"])
        assert res is None

        res = scw.update_user_group(54, None, ["85", "49"])
        assert res is None

    @responses.activate
    def test_delete_user_or_group(self, user_or_group_mock, scw):
        res = scw.delete_user_or_group(70)
        assert res is None

    @responses.activate
    def test_add_user(self, user_or_group_mock, user_mock, scw):
        new_user = NewLocalUser()
        new_user.name = "abaker"
        new_user.first_name = "Aaron"
        new_user.last_name = "Baker"
        new_user.email = "aaron.baker@contoso.com"
        new_user.password = "password"
        new_user.notes = "asdf"
        new_user.default_authentication = False
        res = scw.add_user(new_user)
        assert res.name == "johnny_smith"

    @responses.activate
    def test_user_import_ldap(self, user_or_group_mock, scw):
        ldap = LDAPParty()
        ldap.ldap_configuration_name = "Contoso LDAP"
        ldap.ldap_dn = "CN=abaker,OU=Users,DC=tcse,DC=net"
        res = scw.user_import_ldap(ldap)
        assert res is None

    @responses.activate
    def test_user_login(self, user_or_group_mock, scw):
        user = LDAPUserSimulateLogin()
        user.name = "abaker"
        ldap_config = LDAPConfiguration()
        ldap_config.name = "Contoso LDAP"
        ldap_config.id = 1
        user.ldap_configuration = ldap_config
        res = scw.user_login(user)
        assert res is None

    @responses.activate
    def test_group_login(self, user_or_group_mock, scw):
        group = LDAPSimulateLogin()
        group.name = "Accountants"
        ldap_config = LDAPConfiguration()
        ldap_config.name = "Contoso LDAP"
        ldap_config.id = 1
        group.ldap_configuration = ldap_config
        res = scw.group_login(group)
        assert res is None



from typing import Optional, List

from pytos2.utils import propify, prop
from pytos2.models import Jsonable


@propify
class SecureChangeAddress(Jsonable):
    ip_address: str = prop("")
    type: str = prop("")


@propify
class Property(Jsonable):
    key: str = prop("")
    value: str = prop("")


@propify
class Properties(Jsonable):
    secure_change_addresses: List[SecureChangeAddress] = prop(
        factory=list, flatify="secure_change_address"
    )
    general_properties: List[Property] = prop(factory=list, flatify="property")



import json
import re

import pytest  # type: ignore
import responses

from pytos2.secureapp import SaAPI
from pytos2.secureapp.entrypoint import Sa

from tests.securechange.conftest import users_mock, scw


@pytest.fixture
def sa_api():
    return SaAPI(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def sa():
    return Sa(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def application_identities_mock():
    application_identities_json = json.load(
        open("tests/secureapp/json/application_identities.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/application_identities",
        json=application_identities_json,
    )


@pytest.fixture
def failure_mock():
    responses.add(
        responses.GET,
        re.compile("https://198.18.0.1/securechangeworkflow/api/secureapp.*"),
        status=500,
    )


@pytest.fixture
def applications_mock():
    applications_json = json.load(open("tests/secureapp/json/applications.json"))
    application_54_json = json.load(open("tests/secureapp/json/application_54.json"))

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications",
        json=applications_json,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/54",
        json=application_54_json,
    )


@pytest.fixture
def application_connections_mock():
    application_connections_json = json.load(
        open("tests/secureapp/json/app_60_connections.json")
    )

    application_connection_126_json = json.load(
        open("tests/secureapp/json/app_60_connection_126.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections",
        json=application_connections_json,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/126",
        json=application_connection_126_json,
    )


@pytest.fixture
def application_history_mock():
    application_history_json = json.load(
        open("tests/secureapp/json/app_60_history.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/history",
        json=application_history_json,
    )



from datetime import datetime
from netaddr import IPAddress
import pytest
from pytos2.api import ApiError
import responses

from pytos2.secureapp.history import (
    HistoryConnectionBaseSnapshot,
    HistoryConnectionDetails,
)
from pytos2.securetrack.service_object import TCPServiceObject

from . import conftest  # noqa

from pytos2.secureapp.entrypoint import Sa
from pytos2.secureapp.application_identities import ApplicationIdentity
from dateutil.tz import tzoffset


class TestEntrypoint:
    @responses.activate
    def test_application_identities(self, sa: Sa, application_identities_mock):
        # print(dir(sa))
        identity = sa.application_identities[0]
        assert isinstance(identity, ApplicationIdentity)

    @responses.activate
    def test_application_identites_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.application_identities

    @responses.activate
    def test_applications(self, sa: Sa, applications_mock):
        applications = sa.get_applications()
        assert applications[0].id == 60
        assert (
            applications[0].connections[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/126"
        )
        assert applications[0].owner and applications[0].owner.name == "r"

    @responses.activate
    def test_application(self, sa: Sa, applications_mock):
        application = sa.get_application(54)
        assert application.id == 54
        assert application.owner.name == "Jessica.Sanchez"
        assert (
            application.open_tickets[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/35"
        )

    @responses.activate
    def test_update_application(self, sa: Sa, applications_mock):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/54",
            status=200,
            body="",
        )

        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "There is no application with the specified ID 1337.",
                }
            },
        )

        application = sa.get_application(54)
        application.name = "New Name"

        sa.update_application(application)
        with pytest.raises(ApiError):
            sa.update_application(1337, name="New Name")

        sa.update_application(54, name="New Name", owner=7)

    @responses.activate
    def test_create_application(self, scw, sa: Sa, applications_mock, users_mock):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechange/api/secureapp/repository/applications/54"
            },
        )

        application = sa.add_application(
            name="VPN users access to RnD users", owner="r"
        )

        # Input and output don't really matter because we're mocking the response
        assert application.id == 54
        assert application.owner.name == "Jessica.Sanchez"
        assert application.name == "VPN users access to RnD users"

    @responses.activate
    def test_delete_application(self, sa: Sa, applications_mock):
        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/54",
            status=200,
            body="",
        )

        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "There is no application with the specified ID 1337.",
                }
            },
        )

        sa.delete_application(54)

        with pytest.raises(ApiError):
            sa.delete_application(1337)

    @responses.activate
    def test_bulk_update_applications(self, sa: Sa, applications_mock):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/",
            status=200,
            body="",
        )

        app = sa.get_application(54)

        sa.bulk_update_applications([app])

    @responses.activate
    def test_application_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.get_application(54)

        with pytest.raises(ValueError):
            sa.get_applications()

    @responses.activate
    def test_application_connections(self, sa: Sa, application_connections_mock):
        connections = sa.get_application_connections(60)
        assert connections[0].id == 126
        assert connections[0].name == "Connection 1"
        assert connections[0].sources[0].name == "CRM_01"

    @responses.activate
    def test_application_connection(self, sa: Sa, application_connections_mock):
        connection = sa.get_application_connection(60, 126)
        assert connection.name == "Connection 1"
        assert connection.sources[0].name == "CRM_01"

    @responses.activate
    def test_application_connection_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.get_application_connection(60, 126)

        with pytest.raises(ValueError):
            sa.get_application_connections(60)

    @responses.activate
    def test_application_history(self, sa: Sa, application_history_mock):
        history = sa.get_application_history(60)

        assert history[0].change_description == "Connection deleted"

        assert history[0].change_details.xsi_type.value == "historyConnectionDetailsDTO"
        assert isinstance(history[0].change_details, HistoryConnectionDetails)

        assert history[0].change_details.removed_sources[0].name == "Any"
        assert history[0].change_details.removed_sources[0].id == 1
        assert history[0].change_details.removed_sources[0].display_name == "Any"
        assert (
            history[0].change_details.removed_sources[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/network_objects/1"
        )

        assert history[0].change_details.removed_destinations[0].name == "CRM_01"
        assert history[0].change_details.removed_destinations[0].id == 762
        assert (
            history[0].change_details.removed_destinations[0].display_name == "CRM_01"
        )
        assert (
            history[0].change_details.removed_destinations[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/network_objects/762"
        )

        assert history[0].change_details.removed_services[0].name == "amitay2"
        assert history[0].change_details.removed_services[0].id == 235
        assert history[0].change_details.removed_services[0].display_name == "amitay2"
        assert (
            history[0].change_details.removed_services[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/services/235"
        )

        assert history[0].date == datetime(
            2024, 4, 15, 3, 26, 57, 257000, tzinfo=tzoffset(None, -25200)
        )
        assert history[0].modified_object.display_name == "Connection 2"
        assert history[0].modified_object.id == 203
        assert (
            history[0].modified_object.link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/203"
        )
        assert history[0].modified_object.name == "Connection 2"
        assert history[0].snapshot.comment == ""
        assert history[0].snapshot.id == 203
        assert history[0].snapshot.name == "Connection 2"

        assert isinstance(history[0].snapshot, HistoryConnectionBaseSnapshot)
        assert history[0].snapshot.xsi_type.value == "historyConnectionDTO"

        assert history[0].type == "Connection"
        assert history[0].user.display_name == "Henry Carr"
        assert history[0].user.id == 4
        assert (
            history[0].user.link.href
            == "https://198.18.0.1/securechangeworkflow/api/securechange/users/4"
        )
        assert history[0].user.name == "Henry Carr"

        assert "Connection changed" in history[3].change_description
        assert history[3].snapshot.id == 236
        assert history[3].snapshot.name == "Connection 3"
        assert history[3].snapshot.services[0].name == "http"
        assert isinstance(history[3].snapshot.services[0], TCPServiceObject)
        assert history[3].snapshot.services[0].min_port == 80
        assert history[3].snapshot.services[0].max_port == 80
        assert history[3].snapshot.services[0].protocol == 6

        assert history[3].snapshot.destinations[0].name == "CRM_01"
        assert history[3].snapshot.destinations[0].ip == IPAddress("192.168.205.33")

        # Covering empty case in find_in_detail_list
        assert history[4].change_details.removed_sources == []

        assert history[4].change_details.added_destinations[0].id == 3245
        assert history[4].change_details.added_destinations[0].name == "EDL_server2"
        assert (
            history[4].change_details.added_destinations[0].display_name
            == "EDL_server2"
        )

        assert history[5].change_details.added_services[0].id == 233
        assert history[5].change_details.added_services[0].name == "amitay5"

        assert history[11].change_details.added_sources[0].id == 3244
        assert history[11].change_details.added_sources[0].name == "EDL_subnet"

    @responses.activate
    def test_application_history_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.get_application_history(60)
